#ifndef __JABYENGINE_GPU_TYPES_HPP__
#define __JABYENGINE_GPU_TYPES_HPP__
#include "../Auxiliary/bits.hpp"
#include "../jabyengine_defines.h"

namespace JabyEngine {
    namespace GPU {
        struct Color24 {
            uint8_t red   = 0;
            uint8_t green = 0;
            uint8_t blue  = 0;        

            constexpr Color24() = default;
            constexpr Color24(uint8_t r, uint8_t g, uint8_t b) : blue(b), green(g), red(r) {
            }

            constexpr uint32_t raw() const {
                return ((this->blue << 16) | (this->green << 8) | this->red);
            }

            static constexpr Color24 Black() {
                return Color24(0, 0, 0);
            }

            static constexpr Color24 White() {
                return Color24(0xFF, 0xFF, 0xFF);
            }

            static constexpr Color24 Red() {
                return Color24(0xFF, 0x0, 0x0);
            }

            static constexpr Color24 Green() {
                return Color24(0x0, 0xFF, 0x0);
            }

            static constexpr Color24 Blue() {
                return Color24(0x0, 0x0, 0xFF);
            }
        };

        class Color {
        private:
            static constexpr auto RedRange            = BitRange::from_to(0, 4);
            static constexpr auto GreenRange          = BitRange::from_to(5, 9);
            static constexpr auto BlueRange           = BitRange::from_to(10, 14);
            static constexpr auto SemiTransperancyBit = Bit(15);

            uint16_t value = 0;

        public:
            static constexpr Color from_rgb(uint8_t r, uint8_t g, uint8_t b) {
                return Color().set_red(r).set_green(g).set_blue(b);
            }

            static constexpr Color from(const Color24& color) {
                return Color::from_rgb(color.red, color.green, color.blue);
            }

            constexpr Color& set_red(uint8_t red) {
                this->value = bit::value::set_normalized(this->value, RedRange.with(red));
                return *this;
            }

            constexpr Color& set_green(uint8_t green) {
                this->value = bit::value::set_normalized(this->value, GreenRange.with(green));
                return *this;
            }

            constexpr Color& set_blue(uint8_t blue) {
                this->value = bit::value::set_normalized(this->value, BlueRange.with(blue));
                return *this;
            }
        };

        template<typename T>
        struct Position {
            T x = 0;
            T y = 0;

            constexpr Position() = default;
            constexpr Position(T x, T y) : x(x), y(y) {
            }
        };

        template<typename T>
        struct Size {
            T width  = 0;
            T height = 0;

            constexpr Size() = default;
            constexpr Size(T w, T h) : width(w), height(h) {
            }
        };

        template<typename T>
        struct Area {
            Position<T> position;
            Size<T>     size;

            constexpr Area() = default;
            constexpr Area(Position<T> position, Size<T> size) : position(position), size(size) {
            }
            constexpr Area(T position_x, T position_y, T size_width, T size_height) : position{position_x, position_y}, size{size_width, size_height} {
            }
        };

        // Type used for primitives
        struct PagePosition {
            uint8_t u = 0;
            uint8_t v = 0;

            constexpr PagePosition() = default;
            constexpr PagePosition(uint8_t u, uint8_t v) : u(u), v(v) {
            }
        };

        typedef Position<int16_t>  PositionI16;
        typedef Position<uint16_t> PositionU16;

        typedef Size<int16_t>  SizeI16;
        typedef Size<uint16_t> SizeU16;

        typedef Area<int16_t>  AreaI16;
        typedef Area<uint16_t> AreaU16;

        // Type used for primitives
        typedef PositionI16 Vertex;
    }
}
#endif //!__JABYENGINE_GPU_TYPES_HPP__